@@ -109,9 +109,10 @@ class builder():
         self.string_u2z = ''
         self.u2z = '\#'
 
-        if not os.path.exists('__pycache__'):
-            os.makedirs('__pycache__')
+        if not os.path.exists('build'):
+            os.makedirs('build')
 
+        self.matrices_folder = 'build'
 
 
     def check_system(self):
@@ -1007,8 +1008,11 @@ class builder():
                 source_de_uz += source_de_jac
                 defs_sp_uz   += defs_sp_jac
                 source_sp_uz += source_sp_jac
-
-                save_npz(f"./{sys['name']}_{item}_num.npz",sp_jac_num_matrix, compressed=True)
+                
+                if self.matrices_folder == '':
+                    save_npz(f"./{sys['name']}_{item}_num.npz",sp_jac_num_matrix, compressed=True)
+                else:
+                    save_npz(f"./{self.matrices_folder}/{sys['name']}_{item}_num.npz",sp_jac_num_matrix, compressed=True)
 
 
         ## C sources ##############################################################
@@ -1030,9 +1034,9 @@ class builder():
         source += source_de_trap + source_sp_trap
         source += source_de_uz + source_sp_uz
 
-        save_npz( f"./{sys['name']}_sp_jac_ini_num.npz", sp_jac_ini_num_matrix, compressed=True)
-        save_npz( f"./{sys['name']}_sp_jac_run_num.npz", sp_jac_run_num_matrix, compressed=True)
-        save_npz( f"./{sys['name']}_sp_jac_trap_num.npz",sp_jac_trap_num_matrix, compressed=True)
+        save_npz( f"./{self.matrices_folder}/{sys['name']}_sp_jac_ini_num.npz", sp_jac_ini_num_matrix, compressed=True)
+        save_npz( f"./{self.matrices_folder}/{sys['name']}_sp_jac_run_num.npz", sp_jac_run_num_matrix, compressed=True)
+        save_npz( f"./{self.matrices_folder}/{sys['name']}_sp_jac_trap_num.npz",sp_jac_trap_num_matrix, compressed=True)
 
         if self.verbose: print(f'Code wrote in {time.time()-t_0:0.3f} s')
 
@@ -1059,9 +1063,9 @@ class builder():
         ffi = cffi.FFI()
         name = self.name
 
-        with open(f'./__pycache__/defs_{name}_cffi.h', 'r') as f:
+        with open(f'./build/defs_{name}_cffi.h', 'r') as f:
             defs = f.read()
-        with open(f'./__pycache__/source_{name}_cffi.c', 'r') as f:
+        with open(f'./build/source_{name}_cffi.c', 'r') as f:
             source = f.read()
             
         ffi.cdef(defs, override=True)
@@ -1078,9 +1082,9 @@ class builder():
         self.jacobians()
         defs,source = self.sym2src()
         if self.save_sources:
-            with open(f'./__pycache__/defs_{name}_cffi.h', 'w') as f:
+            with open(f'./build/defs_{name}_cffi.h', 'w') as f:
                 f.write(defs)
-            with open(f'./__pycache__/source_{name}_cffi.c', 'w') as f:
+            with open(f'./build/source_{name}_cffi.c', 'w') as f:
                 f.write(source)
         self.compile_module()
         self.sys2num()
