@@ -10,8 +10,11 @@ import numpy as np
 import pandas as pd
 import scipy
 import matplotlib.pyplot as plt
+from scipy.sparse.linalg import spsolve
+from scipy.sparse import csc_matrix
 
 from matplotlib.patches import Circle, Wedge, Polygon, Rectangle
+import pydae.ctrl as ctrl
 
 def eval_A(system):
     
@@ -101,6 +104,10 @@ def eval_ss(system):
     return A
 
 
+def ss_eval(model):
+    return ctrl.ss_eval(model)
+
+
 
 def eval_A_ini(system):
     
@@ -116,8 +123,13 @@ def eval_A_ini(system):
     return A
 
 
-def damp_report(system):
-    eig,eigv = np.linalg.eig(system.A)
+def damp_report(system, sparse=False):
+    
+    if sparse:
+        eig,eigv = np.linalg.eig(system.A.toarray())
+    else:      
+        eig,eigv = np.linalg.eig(system.A)
+        
     omegas = eig.imag
     sigmas = eig.real
 
@@ -153,6 +165,47 @@ def damp_report(system):
     system.eigvectors = eigv
     return eig_df
  
+def damp(A, sparse=False):
+    
+    if sparse:
+        eig,eigv = np.linalg.eig(A)
+    else:      
+        eig,eigv = np.linalg.eig(A)
+        
+    omegas = eig.imag
+    sigmas = eig.real
+
+    freqs = np.abs(omegas/(2*np.pi))
+    zetas = -sigmas/np.sqrt(sigmas**2+omegas**2)
+    
+    string = ''
+    string += f' Mode'.ljust(10, ' ') 
+    string += f' Real'.ljust(10, ' ') 
+    string += f' Imag'.ljust(10, ' ') 
+    string += f' Freq.'.ljust(10, ' ') 
+    string += f' Damp'.ljust(10, ' ') 
+    string += '\n'
+
+    N_x = len(eig)
+    for it in range(N_x):
+        r = eig[it].real
+        i = eig[it].imag
+        string += f'{it+1:0d}  '.rjust(10, ' ') 
+        string += f'{r:0.4f}  '.rjust(10, ' ') 
+        string += f'{i:0.4f}j'.rjust(10, ' ') 
+        string += f'{freqs[it]:0.4f}'.rjust(10, ' ') 
+        string += f'{zetas[it]:0.4f}'.rjust(10, ' ') 
+
+        string += '\n'
+        #'\t{i:0.4f}\t{freqs[it]:0.3f}\t{zetas[it]:0.4f}\n'
+    
+    columns = ['Real','Imag','Freq.','Damp']     
+    modes = [f'Mode {it+1}' for it in range(N_x)]
+    eig_df = pd.DataFrame(data={'Real':eig.real,'Imag':eig.imag,  'Freq.':freqs,     'Damp':zetas},index=modes)
+    
+
+    return eig_df
+
 
 def participation(system, method='kundur'):
     '''
@@ -276,9 +329,18 @@ def lqr(A,B,Q,R):
 
 
 
-def plot_eig(grid, x_min='',x_max='',y_min='',y_max=''):
-    fig,axes = plt.subplots()
+def plot_eig(eigenvalues, x_min='',x_max='',y_min='',y_max='', fig='', mark='o'):
+    ''''
+    Creates a matplotlib figure from a numpy array of eigenvalues.
 
+
+    
+    '''
+
+    if fig == '':
+        fig,axes = plt.subplots()
+    else:
+        axes = fig.axes[0]
     
     pi2 = 2*np.pi
 
@@ -308,14 +370,12 @@ def plot_eig(grid, x_min='',x_max='',y_min='',y_max=''):
     axes.add_patch(poly)
     #axes.text(-0.37, 9.0/pi2, '$\zeta<$5 %', fontsize=12)
 
-
-
     axes.plot([0,0],[-20,20],'-', color='k', lw=4)
 
-    if x_min == '': x_min = np.min(grid.eigvalues.real)*1.1
-    if x_max == '': x_max = np.max(grid.eigvalues.real)
-    if y_min == '': y_min = np.min(grid.eigvalues.imag)/(2*np.pi)*1.1
-    if y_max == '': y_max = np.max(grid.eigvalues.imag)/(2*np.pi)*1.1
+    if x_min == '': x_min = np.min(eigenvalues.real)*1.1
+    if x_max == '': x_max = np.max(eigenvalues.real)
+    if y_min == '': y_min = np.min(eigenvalues.imag)/(2*np.pi)*1.1
+    if y_max == '': y_max = np.max(eigenvalues.imag)/(2*np.pi)*1.1
         
     axes.set_xlim((x_min,x_max))
     axes.set_ylim((y_min,y_max))
@@ -323,10 +383,10 @@ def plot_eig(grid, x_min='',x_max='',y_min='',y_max=''):
     axes.set_xlabel('Real')
     axes.set_ylabel('Imag$/2\pi$ (Hz)')
 
-    axes.plot(grid.eigvalues.real,grid.eigvalues.imag/(2*np.pi),'o')
+    axes.plot(eigenvalues.real,eigenvalues.imag/(2*np.pi),mark)
     fig.tight_layout()
     
-    return fig
+    return fig;
     
 def add_arrow(string,name='arrow',scale=1,angle=0,center_x=0,center_y=0,color="#337ab7"):
     trans_x = center_x*(1-scale)
